! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!+
! Module: omgrismresp_mod
!
! Authors:  Richard Saxton - SOC
!          
! Description: Prepares OM grism files for spectral fitting
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module omgrismresp_mod

  use dal
  use param  
  use errorHandling

  implicit none 
  
  private 

  public omgrismresp

contains 
  
  !---------------------------------------------------------------------
  subroutine omgrismresp
  !---------------------------------------------------------------------
      character*80 inspec
      character*80 comment,dummy
      character*60 outspec, respfile

      character*20 filter
      character*255 sas_dir,omgrismcaldir
      character*276 Vcalfile,UVcalfile
      integer blksize, istat,hdutype, nrow, cnum
      integer chancnt,cnt,ncol,encount
      integer maxlen
         parameter (maxlen=10000)
      real wave(maxlen),rate(maxlen),error(maxlen)
      real cpersec(maxlen),cpersec_err(maxlen)
      real enlow(maxlen),enhigh(maxlen)
      real width,exposure
      integer(kind=int16) chan(maxlen),nchan(maxlen),fchan(maxlen),ngrp(maxlen)

      integer lp,elp,clp
      integer tfields
      logical anyf
      character*10 ttype(6),tform(6),tunit(6),extname
      real resp(maxlen),matrix(maxlen),chanwave(maxlen),expbit
      real sum,newsum,testwave,matwave(maxlen)
      real area_wave(maxlen),areas(maxlen)
      integer nareas
   
! Set the sigma for the Gaussian line spread function.
! FWHM of 15 Angstrom for the UV grism and 25 for the V grism
      real sigma,sigma_v,sigma_uv
      DATA sigma_v/10.62/
      DATA sigma_uv/6.36/

! Set the threshold for including response elements
      real threshold
      DATA threshold/1.0E-6/

! Limits for spectral extraction
      real vmin, vmax, uvmin, uvmax, min_wave, max_wave
        parameter (vmin=3000.0)
        parameter (vmax=6000.0)
        parameter (uvmin=1800.0)
        parameter (uvmax=3600.0)

      istat=0
!
!  Open the input SAS spectrum
      inspec = trim(stringParameter("inputspec"))
      call ftopen(1, inspec, 0, blksize, istat)

      if (istat .ne. 0) then
         write(*,*)' Error: failed to open Input spectrum ',inspec
         goto 100
      endif

!  Open the output spectrum
      outspec = trim(stringParameter("outputspec"))
      call ftinit(2, outspec, blksize, istat)

      if (istat .ne. 0) then
         write(*,*)' WriteError: failed to open output spectrum ',outspec,&
                   ' perhaps this file already exists?'
         goto 100
      endif

!  Open the output response file
      respfile = trim(stringParameter("responsefile"))
      call ftinit(3, respfile, blksize, istat)

      if (istat .ne. 0) then
         write(*,*)' WriteError: failed to open output resp file ',respfile,&
                   ' perhaps this file already exists?'
         goto 100
      endif

! Read the filter keyword from the input spectrum
      call ftgkys(1,'FILTER',filter,comment,istat)

      if (istat .ne. 0) then
         write(*,*)' Error: failed to read filter name from spectrum'
         goto 100
      endif

! Get the location of the Grism CAL files
      call get_environment_variable('SAS_DIR',sas_dir)
      omgrismcaldir = trim(sas_dir) // '/lib/data/omgrismresp_data'
      UVcalfile=TRIM(omgrismcaldir) // '/OM_UVgrism_efArea.txt'
      Vcalfile=TRIM(omgrismcaldir) // '/OM_Vgrism_efArea.txt'

! Set the sigma of the line spread function and the wavelength range 
      if (filter .eq. 'GRISM1') then
         min_wave=uvmin
         max_wave=uvmax
         sigma=sigma_UV
         open(unit=4,file=TRIM(UVcalfile),status='old') 
      else if (filter .eq. 'GRISM2') then
         min_wave=vmin
         max_wave=vmax
         sigma=sigma_V
         open(unit=4,file=TRIM(Vcalfile),status='old')
      else
         write(*,*)' Error: Filter ',filter,' not recognised'
         goto 100
      endif

! Read in the relevant effective area file - first 3 records are text
      do lp=1,3
         read(4,*)dummy
      enddo
      do lp=1,10000
         read(4,*,end=10)area_wave(lp),areas(lp)
      enddo

10    continue

      nareas=lp-1

! Move to the spectrum header
      call ftmahd(1, 2, hdutype, istat)

      if (istat .ne. 0) then
         write(*,*)' Error: failed to move to spectrum extension'
         goto 100
      endif

! Read the wavelength, rate and error columns
      call ftgncl(1, ncol, istat)
      call ftgnrw(1, nrow, istat)

      call ftgcno(1, 0, 'WAVELENGTH', cnum, istat)
      call ftgcve(1, cnum, 1, 1, nrow, 0, wave, anyf, istat)
      call ftgcno(1, 0, 'RATE', cnum, istat)
      call ftgcve(1, cnum, 1, 1, nrow, 0, rate, anyf, istat)
      call ftgcno(1, 0, 'ERROR', cnum, istat)
      call ftgcve(1, cnum, 1, 1, nrow, 0, error, anyf, istat)

      if (istat .ne. 0) then
         write(*,*)' Error: failed to read data from input spectrum'
         goto 100
      endif

! Loop over input rows and set the output channels and count rate arrays
      chancnt=0
      do lp=nrow,1,-1

!         write(*,*)lp,wave(lp)

!      Use this data point if the wavelength is within the required
!      range (we assume that rows are ordered in wavelength)

         if (wave(lp) .ge. min_wave .and. wave(lp) .lt. max_wave) then

          if (lp .lt. nrow) then  ! If not the final record
            width=wave(lp+1)-wave(lp)
          else
            width=wave(lp)-wave(lp-1)
          endif

!       Set the output channel
          chancnt=chancnt+1
          chan(chancnt)=chancnt

!       Set the output rate and error
          cpersec(chancnt) = rate(lp) * width
          cpersec_err(chancnt) = error(lp) * width

!       Set the channel energies (in keV) from wavelength in Angstrom
          enlow(chancnt) = 12.3997 / (wave(lp)+width/2.0) 
          enhigh(chancnt) = 12.3997 / (wave(lp)-width/2.0)
          chanwave(chancnt) = wave(lp)

!          write(*,*)chan(chancnt),cpersec(chancnt),cpersec_err(chancnt)
         endif
      enddo

! Create the output spectrum 
      tfields=3
      ttype(1)='CHANNEL'
      ttype(2)='RATE'
      ttype(3)='STAT_ERR'
      tform(1)='I'
      tform(2)='E'
      tform(3)='E'
      tunit(1)=' '
      tunit(2)='counts/s'
      tunit(3)='counts/s'
      extname='SPECTRUM'

! Update the channel column
      call ftibin(2,chancnt,tfields,ttype,tform,tunit,extname,0,istat)
      call FTPCLI(2,1,1,1,chancnt,chan,istat)
      call FTPCLE(2,2,1,1,chancnt,cpersec,istat)
      call FTPCLE(2,3,1,1,chancnt,cpersec_err,istat)
 
! Read the exposure keyword from the input spectrum
      call FTGKYE(1,'EXPOSURE',exposure,comment,istat)

! Append header keywords to the output spectrum
      call addspechead(2,chancnt,respfile,exposure,istat)

      if (istat .ne. 0) then
         write(*,*)' Error: writing to output spectrum'
         goto 100
      endif

! Close the output spectrum
      call ftclos(1, istat)
      call ftclos(2, istat)

! Make the extension for the response matrix
      tfields=6
      ttype(1)='ENERG_LO'
      ttype(2)='ENERG_HI'
      ttype(3)='N_GRP'
      ttype(4)='F_CHAN'
      ttype(5)='N_CHAN'
      ttype(6)='MATRIX'
      tform(1)='E'
      tform(2)='E'
      tform(3)='I'
      tform(4)='PI'
      tform(5)='PI'
      tform(6)='PE'
      tunit(1)='keV'
      tunit(2)='keV'
      tunit(3)=' '
      tunit(4)=' '
      tunit(5)=' '
      tunit(6)=' '
      extname='MATRIX'

! Initially we use the same test energies in the RMF as the channel energies
      encount=chancnt
      call ftibin(3,encount,tfields,ttype,tform,tunit,extname,0,istat)

! Set the energies - for starters use the same as the channel energies this
! can probably be reduced with experience
      call FTPCLE(3,1,1,1,encount,enlow,istat)
      call FTPCLE(3,2,1,1,encount,enhigh,istat)

!   Loop over each energy
      do elp=1,encount
         sum=0.0

         testwave=chanwave(elp) ! Set the same for now

!      Loop over each channel
         do clp=1,chancnt

!         Calculate the distribution of this energy in this channel 
            expbit = (chanwave(clp) - testwave) **2
            expbit = expbit / (2.0 * sigma * sigma)
            resp(clp)=exp(-expbit)
            sum=sum+resp(clp)
!            write(*,*)' RESP: ',clp,resp(clp),expbit
         enddo

!      Normalise and threshold
         fchan(elp)=0
         cnt=0
         newsum=0.0

         do clp=1,chancnt
            resp(clp)=resp(clp)/sum

!            write(*,*)'En, resp, thresh: ',clp,resp(clp),threshold
            if (resp(clp) .ge. threshold) then

!            Set start channel 
               if (fchan(elp) .eq. 0) then
                  fchan(elp)=clp
               endif

!            Add one to total number of above threshold channels
               cnt=cnt+1

               matrix(cnt)=resp(clp)
               matwave(cnt)=wave(clp)
               newsum=newsum+resp(clp)
            endif
         enddo

!      Loop over output channels
         do lp=1,cnt

!         Final normalisation of output values
            matrix(lp)=matrix(lp)/newsum
         enddo

!      Set the number of groups and channels array
         ngrp(elp)=1
         nchan(elp)=cnt

         call FTPCLI(3,4,elp,1,1,fchan(elp),istat)
         call FTPCLI(3,5,elp,1,1,nchan(elp),istat)

!      Finally multiply the matrix by the effective area
         call effarea(maxlen,nareas,areas,area_wave,testwave,matrix) 
         call FTPCLE(3,6,elp,1,cnt,matrix,istat)

      enddo

! Set the Ngrp array in one go
      call FTPCLI(3,3,1,1,encount,ngrp,istat)

! Add keywords to this extension
      call addresphead(3,chancnt,istat)

      if (istat .ne. 0) then
         write(*,*)' Error: writing to output response matrix'
         goto 100
      endif

! Now write the ebounds extension
      tfields=3
      ttype(1)='CHANNEL'
      ttype(2)='E_MIN'
      ttype(3)='E_MAX'
      tform(1)='I'
      tform(2)='E'
      tform(3)='E'
      tunit(1)=' '
      tunit(2)='keV'
      tunit(3)='keV'
      extname='EBOUNDS'

      call ftibin(3,encount,tfields,ttype,tform,tunit,extname,0,istat)
      call FTPCLI(3,1,1,1,chancnt,chan,istat)
      call FTPCLE(3,2,1,1,chancnt,enlow,istat)
      call FTPCLE(3,3,1,1,chancnt,enhigh,istat)

      if (istat .ne. 0) then
         write(*,*)' Error: writing to output response matrix'
         goto 100
      endif

! Close the files
      call ftclos(3, istat)

100   continue


  end subroutine omgrismresp
  
! Subroutine to multiply a redistribution curve by effective areas
      subroutine effarea(maxlen,nareas,areas,area_wave,testwave,matrix)

      implicit none

      integer maxlen,nareas
      real areas(maxlen),area_wave(maxlen),testwave
      real matrix(maxlen)

      integer lp
      real areafact

!   Find the area corresponding to this wave
      if (testwave .lt. area_wave(1) .or. &
          testwave .gt. area_wave(nareas)) then
         areafact=0.0
      else
         do lp=1,nareas-1
            if (testwave .ge. area_wave(lp) .and. &
                 testwave .le. area_wave(lp+1)) then
               
               areafact = areas(lp+1) - &
                          (areas(lp+1)-areas(lp)) * &
                          (area_wave(lp+1)-testwave) / &
                          (area_wave(lp+1)-area_wave(lp)) 
            endif
         enddo
      endif

      do lp=1,maxlen
         matrix(lp)=matrix(lp)*areafact
      enddo

!      write(*,*)'AREA factor: ',areafact,testwave

      end subroutine effarea

! Subroutine to add keywords to spectral header
      subroutine addspechead(unit,detchan,respfile,exposure,istat)

      implicit none

      integer unit,detchan,lp
      character*80 card(18),detcard
      character*60 respfile
      real exposure
      integer istat
      DATA card/ &
      "HDUCLASS= 'OGIP  ' / format conforms to OGIP/GSFC standards",    &
      "HDUCLAS1= 'SPECTRUM'           / Extension contains a Spectrum", &
      "HDUVERS1= '1.1.0   '           / Version number of the format", &
      "HDUCLAS2= 'NET'                / Bckgnd has been subtracted", &
      "HDUCLAS3= 'RATE'               / Exposure corrected", &
      "TELESCOP= 'XMM     '", &
      "INSTRUME= 'OM      '           / EPIC PN instrument", &
      "DETNAM  = 'OM1      '          / specific detector name in use", &
      "CHANTYPE= 'PI      '           / Detector Channel Type in use", &
      "BACKSCAL= 1.0     ", &
      "CORRSCAL= 1.0     ", &
      "AREASCAL= 1.0     ", &
      "GROUPING= 0       ", &
      "QUALITY = 0       ", &
      "TLMIN   = 1       ", &
      "BACKFILE= 'NONE' ",  &
      "ANCRFILE= 'NONE' ",  &
      "POISSERR= F"/

!   Reserve space to write keywords
      call FTHDEF(unit,40,istat)

!   Write keywords
      do lp=1,18
         call FTPREC(unit,card(lp),istat)
      enddo

      write(detcard,fmt='(A10,I10)')"DETCHANS= ",detchan
      call FTPREC(unit,detcard,istat)

      write(detcard,fmt='(A10,F10.1)')"EXPOSURE= ",exposure
      call FTPREC(unit,detcard,istat)

      detcard="RESPFILE= '" // respfile // "'"
      call FTPREC(unit,detcard,istat)

      end subroutine addspechead

! Subroutine to add keywords to response header
      subroutine addresphead(unit,detchan,istat)

      implicit none

      integer unit,detchan,lp
      character*80 card(11),detcard
      integer istat
      DATA card/ &
      "HDUCLASS= 'OGIP  ' / format conforms to OGIP/GSFC standards", &
      "HDUCLAS1= 'RESPONSE'           / Extension contains a Spectrum", &
      "HDUVERS1= '1.0.0   '           / Version number of the format", &
      "HDUCLAS2= 'RSP_MATRIX'  / Extension contains spectral response", &
      "HDUVERS2= '1.1.0   '           / Version number of the format", &
      "HDUCLAS3= 'DETECTOR'", &
      "TELESCOP= 'XMM     '", &
      "INSTRUME= 'OM      '           / EPIC PN instrument", &
      "DETNAM  = 'OM1      '          / specific detector name in use", &
      "CHANTYPE= 'PI      '           / Detector Channel Type in use", &
      "LO_THRES=              0.000001 / threshold for stored matrix"/ 

!   Reserve space to write keywords
      call FTHDEF(unit,40,istat)

      do lp=1,11
         call FTPREC(unit,card(lp),istat)
      enddo

      write(detcard,fmt='(A11,I10,A1)')"DETCHANS= '",detchan,"'"

      call FTPREC(unit,detcard,istat)

      end subroutine addresphead

end module omgrismresp_mod
